/*
This is a cpython extension module named "impiatexit" written using C++.
It implements the module functions "initialize" and "testmpi".
The "initialize" function calls MPI_Init, and registers a MPI_Finalize call
at program exit using std::atexit.
The "testmpi" function needs MPI to be initialized and simply prints
some MPI ranks in a Hello-World fashion.
*/
#define PY_SSIZE_T_CLEAN
#include "Python.h"
#include "mpi.h"
#include <iostream>
#include <cstdlib>

// Simple checker if MPI function call was successful.
static void checkMPIError(int mpi_error_code)
{
    if (mpi_error_code == MPI_SUCCESS)
    {
        return;
    }

    throw "MPI related call did not return successfully!\n";
}

// Wrapper around MPI_Finalize
static void finalize()
{
    std::cout << "Finalizing MPI...\n";
    checkMPIError(MPI_Finalize());
}

// We use C++ code, but a C python interpreter, so we need to use extern "C" for
// everything exposed in python.
extern "C" {

// Module function initializing MPI and registering it for finalization.
static PyObject *
impiatexit_initialize(PyObject *self, PyObject *args)
{
    if (!PyArg_ParseTuple(args, ":initialize"))
        return NULL;

    std::cout << "Initializing MPI...\n";
    checkMPIError(MPI_Init(static_cast<int*>(NULL), static_cast<char***>(NULL)));

    std::cout << "Registering finalization at exit...\n";
    if (std::atexit(finalize) != 0) {
        std::cerr << "Failed to register atexit call to finalize" << std::endl;
    }

    Py_INCREF(Py_None);
    return Py_None;
}

// A a module function testing MPI without initialization/finalization.
// This the code is taken from the "test.cpp" of IMPI.
static PyObject *
impiatexit_testmpi(PyObject *self, PyObject *args)
{
    if (!PyArg_ParseTuple(args, ":testmpi"))
        return NULL;

    // Copied example for MPI communication.
    int i, rank, size, namelen;
    char name[MPI_MAX_PROCESSOR_NAME];
    MPI::Status stat;

    size = MPI::COMM_WORLD.Get_size();
    rank = MPI::COMM_WORLD.Get_rank();
    MPI::Get_processor_name(name, namelen);

    if (rank == 0) {
        std::cout << "Hello world: rank " << rank << " of " << size << " running on " << name <<
            "\n";

        for (i = 1; i < size; i++) {
            MPI::COMM_WORLD.Recv(&rank, 1, MPI_INT, i, 1, stat);
            MPI::COMM_WORLD.Recv(&size, 1, MPI_INT, i, 1, stat);
            MPI::COMM_WORLD.Recv(&namelen, 1, MPI_INT, i, 1, stat);
            MPI::COMM_WORLD.Recv(name, namelen + 1, MPI_CHAR, i, 1, stat);
            std::cout << "Hello world: rank " << rank << " of " << size << " running on " << name <<
                "\n";
        }
    } else {
        MPI::COMM_WORLD.Send(&rank, 1, MPI_INT, 0, 1);
        MPI::COMM_WORLD.Send(&size, 1, MPI_INT, 0, 1);
        MPI::COMM_WORLD.Send(&namelen, 1, MPI_INT, 0, 1);
        MPI::COMM_WORLD.Send(name, namelen + 1, MPI_CHAR, 0, 1);
    }

    // Return None
    Py_INCREF(Py_None);
    return Py_None;
}

// The Module's Method Table
static PyMethodDef ImpisimpleMethods[] = {
    {"testmpi", impiatexit_testmpi, METH_VARARGS, "Test mpi."},
    {"initialize", impiatexit_initialize, METH_VARARGS, "MPI init/finit."},
    {NULL, NULL, 0, NULL}
};

// The Module definition structure.
static struct PyModuleDef impiatexitmodule = {
    PyModuleDef_HEAD_INIT,
    "impiatexit",
    NULL,
    -1,
    ImpisimpleMethods
};

// end extern "C"
}

// The module's initialization function. (The macro return type accounts for extern "C")
PyMODINIT_FUNC
PyInit_impiatexit(void)
{
    return PyModule_Create(&impiatexitmodule);
}
